
//***********************
// TLC Layer Mover/Copier
// by d.e.boyce
//***********************

// This is the source code for the MCLAYER.EXE utility program.
// It is included as an example of how to write simple programs
// that work with LASI cells in TLC format.

// Most of the functions are generic and may be used in other programs.
// Compile with Microsoft C/C++ 1.0 or 1.5 as DOS executable in small model. 


typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned long ulong;

#define VERS    "5.2"
#define NAMLEN	8

#include <stdio.h>
#include <math.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#define TLCLIM  100
#define VTXMAX  512
#define CELMAX  512

//function prototypes
int LoadREF(void);
int ropenTLC(char *);
int wopenTLC(char *);
int read4eqCode(void);
int read4Hdr(void);
int readC(void);
int readB(void);
int readP(void);
int readT(void);
void writeH(void);
void writeC(void);
void writeB(void);
void writeP(void);
void writeT(void);
void moveP(void);
void moveB(void);
void moveT(void);
void moveC(void);
void copyP(void);
void copyB(void);
void copyT(void);
void copyC(void);
int abrtCk(void);
char *chop(char *,int);
void Help(void);


//global variables
int Ncells,OldLayer,NewLayer,Action;
char TLCname[CELMAX][NAMLEN+1];
FILE *rTLCfile,*wTLCfile;

//TLC record structs
struct tlchdr
{
    char name[NAMLEN+1];
    char lv[6];
    char tv[6];
    double scale;
    char punit[10];
    char date[9];
    char time[9];
    int rank;
    long l;
    long b;
    long r;
    long t;
    int bn;
    int pn;
	long vn;
    int cn;
} TLCh={"",VERS,VERS,1.,"um",1,0,0,0,0,0,0,0,0};

struct tlccel
{
    char name[NAMLEN+1];
    int o;
    long x;
    long y;
    int r;
} TLCc;

struct tlcbox
{
    int lyr;
    long l;
    long b;
    long r;
    long t;
} TLCb;

struct tlcpath
{
    int lyr;
    long w;
	int nv;
    long x[VTXMAX];
    long y[VTXMAX];
} TLCp;

struct tlctext
{
    int lyr;
    long siz;
	int nv;
    int o;
    long x;
    long y;
	char text[81];
} TLCt;



//*************
//start of main
//*************

main(int argc,char *argv[])
{
    int p=0;
    static char copybuf[TLCLIM];

    printf("\nLASI TLC Layer Mover/Copier  Version 5.2\n\n");

    if(argc<4){
        Help();
    	exit(0);
    }
    
    if(!stricmp(argv[1],"m")) Action=1;
    if(!stricmp(argv[1],"c")) Action=2;
     
    if(!Action){
    	Help();
    	exit(0);

    }
         
    OldLayer=atoi(argv[2]);
    if(OldLayer<1 || OldLayer>64){
        printf("Old Layer out of Range!\n");
    	exit(0);
    }
    
    NewLayer=atoi(argv[3]);
    if(NewLayer<1 || NewLayer>64){
        exit(0);
    }
    
    if((Ncells=LoadREF())<=0)
    	exit(0);

    printf("%d Cells\n",Ncells);

    //cell loop
    for(p=1;p<=Ncells;p++){
        if(!ropenTLC(TLCname[p]))
            continue;
        if(!wopenTLC("TMPTLC"))
            continue;

        if(read4Hdr())
            writeH();

        //read again for objects
        rewind(rTLCfile);
        read4eqCode();
        fclose(rTLCfile);
        fclose(wTLCfile);

        //copy new file to old
        if(!ropenTLC("TMPTLC"))
            continue;
        if(!wopenTLC(TLCname[p]))
            continue;

        do{
            fgets(copybuf,TLCLIM,rTLCfile);
            if(!feof(rTLCfile))
                fputs(copybuf,wTLCfile);
        }while(!feof(rTLCfile));
        fclose(rTLCfile);
        fclose(wTLCfile);
        
        switch(Action){
        case 1:
        	printf("Layer %d Moved to %d in %s.TLC!\n",
        			OldLayer,NewLayer,TLCname[p]);
        	break;
        case 2:
        	printf("Layer %d Copied to %d in %s.TLC!\n",
        			OldLayer,NewLayer,TLCname[p]);
        }
        if(abrtCk()) break;
    }
    return 0;
}

//******************
//start of functions
//******************

//read TLC header info
int read4Hdr(void)
{
    static char buf[TLCLIM];
    char *s;

    buf[0]=0;
    while(!feof(rTLCfile)){
		if(!strcmp(fgets(buf,TLCLIM,rTLCfile),"=H\n")){
            fgets(buf,TLCLIM,rTLCfile);
            strcpy(TLCh.name,chop(buf,NAMLEN));

            fgets(buf,TLCLIM,rTLCfile);
            strcpy(TLCh.lv,chop(buf,5));

            fgets(buf,TLCLIM,rTLCfile);
            strcpy(TLCh.tv,chop(buf,5));

            fgets(buf,TLCLIM,rTLCfile);
            TLCh.scale=atof(buf);

            fgets(buf,TLCLIM,rTLCfile);
            strcpy(TLCh.punit,chop(buf,5));

            fgets(buf,TLCLIM,rTLCfile);
            strcpy(TLCh.date,chop(buf,8));

            fgets(buf,TLCLIM,rTLCfile);
            strcpy(TLCh.time,chop(buf,8));

            fgets(buf,TLCLIM,rTLCfile); //outline
            chop(buf,36);
            if((s=strtok(buf," "))==NULL) return 0;
            TLCh.rank=atoi(s);
            if((s=strtok(NULL," "))==NULL) return 0;
            TLCh.l=atol(s);
            if((s=strtok(NULL," "))==NULL) return 0;
            TLCh.b=atol(s);
            if((s=strtok(NULL," "))==NULL) return 0;
            TLCh.r=atol(s);
            if((s=strtok(NULL," "))==NULL) return 0;
            TLCh.t=atol(s);

            fgets(buf,TLCLIM,rTLCfile); //count
            chop(buf,30);
            if((s=strtok(buf," "))==NULL) return 0;
            TLCh.bn=atoi(s);
            if((s=strtok(NULL," "))==NULL) return 0;
            TLCh.pn=atoi(s);
            if((s=strtok(NULL," "))==NULL) return 0;
            TLCh.vn=atol(s);
            if((s=strtok(NULL," "))==NULL) return 0;
            TLCh.cn=atoi(s);
        }
    }
    return 1;
}

//read for record types then write new records
int read4eqCode(void)
{
    static char buf[TLCLIM];

    buf[0]=0;

    while(!feof(rTLCfile)){

        fgets(buf,TLCLIM,rTLCfile);

        if(feof(rTLCfile))   //eof check
            break;

		if(!strcmp(buf,"=C\n")){
            if(readC()){
             	switch(Action){
            	case 1:
                	moveC();
                	break;
            	case 2:
            		copyC();
            	}
            }
            else
                printf("  Cell Record Read Error!\n");
            continue;
        }
		if(!strcmp(buf,"=B\n")){
            if(readB()){
             	switch(Action){
            	case 1:
                	moveB();
                	break;
            	case 2:
            		copyB();
            	}
            }
            else
                printf("  Box Record Read Error!\n");
            continue;
        }
		if(!strcmp(buf,"=P\n")){
            if(readP()){
             	switch(Action){
            	case 1:
                	moveP();
                	break;
            	case 2:
            		copyP();
            	}
            }
            else
                printf("  Path Record Read Error!\n");
            continue;
        }
		if(!strcmp(buf,"=T\n")){
            if(readT()){
             	switch(Action){
            	case 1:
                	moveT();
                	break;
            	case 2:
            		copyT();
            	}
            }
            else
                printf("  Text Record Read Error!\n");
            continue;
        }
    }
    return 1;
}

//****************
//move operations
//****************

void moveC(void)
{
    writeC();
}

void moveB(void)
{
    if(TLCb.lyr==OldLayer)
        TLCb.lyr=NewLayer;
    writeB();
}

void moveT(void)
{
    if(TLCt.lyr==OldLayer)
        TLCt.lyr=NewLayer;
    writeT();
}

void moveP(void)
{
    if(TLCp.lyr==OldLayer)
        TLCp.lyr=NewLayer;
    writeP();
}

//****************
//copy operations
//****************

void copyC(void)
{
    writeC();
}

void copyB(void)
{
    writeB();
    if(TLCb.lyr==OldLayer)
        TLCb.lyr=NewLayer;
    writeB();
}

void copyT(void)
{
    writeT();
    if(TLCt.lyr==OldLayer)
        TLCt.lyr=NewLayer;
    writeT();
}

void copyP(void)
{
    writeP();
    if(TLCp.lyr==OldLayer)
        TLCp.lyr=NewLayer;
    writeP();
}


//************************
//TLC readers and writers
//************************

//read a cell record
int readC(void)
{
    static char buf[TLCLIM];
    char *s;

    buf[0]=0;
    if(!feof(rTLCfile)){
        fgets(buf,TLCLIM,rTLCfile);
        strcpy(TLCc.name,chop(buf,NAMLEN+1));

        fgets(buf,TLCLIM,rTLCfile);
        chop(buf,36);

        if((s=strtok(buf," "))==NULL) return 0;
        TLCc.o=atoi(s) & 0x07;

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCc.x=atol(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCc.y=atol(s);
        return 1;
    }
    return 0;
}

//read a box record
int readB(void)
{
    static char buf[TLCLIM];
    char *s;

    buf[0]=0;
    if(!feof(rTLCfile)){
        fgets(buf,TLCLIM,rTLCfile);
        chop(buf,TLCLIM);

        if((s=strtok(buf," "))==NULL) return 0;
        TLCb.lyr=atoi(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCb.l=atol(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCb.b=atol(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCb.r=atol(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCb.t=atol(s);
        return 1;
    }
    return 0;
}

//read a path record
int readP(void)
{
    static char buf[TLCLIM];
    char *s;
    int nv=0,p=0;

    buf[0]=0;
    if(!feof(rTLCfile)){

        fgets(buf,TLCLIM,rTLCfile);
        chop(buf,TLCLIM);

        if((s=strtok(buf," "))==NULL) return 0;
        TLCp.lyr=atoi(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCp.w=atol(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        nv=TLCp.nv=atoi(s);

        fgets(buf,TLCLIM,rTLCfile);
        chop(buf,TLCLIM);

        if((s=strtok(buf," "))==NULL) return 0;
        TLCp.x[p]=atol(s);  //1st X

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCp.y[p]=atol(s);  //1st Y
        p++;
        nv--;
        while(nv>0){
            if((s=strtok(NULL," "))==NULL){ //end of line read new line
                fgets(buf,TLCLIM,rTLCfile);
                chop(buf,TLCLIM);
                if((s=strtok(buf," "))==NULL) return 0;
            }
            TLCp.x[p]=atol(s);  //keep X

            if((s=strtok(NULL," "))==NULL) return 0;
            TLCp.y[p]=atol(s);  //keep Y
            p++;
            nv--;
        }
        return 1;
    }
    return 0;
}

//read a text record
int readT(void)
{
    static char buf[TLCLIM];
    char *s;

    buf[0]=0;
    if(!feof(rTLCfile)){
        fgets(buf,TLCLIM,rTLCfile);
        chop(buf,TLCLIM);

        if((s=strtok(buf," "))==NULL) return 0;
        TLCt.lyr=atoi(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCt.siz=atol(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCt.nv=atoi(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCt.o=atoi(s);

        //coords
        fgets(buf,TLCLIM,rTLCfile);
        chop(buf,TLCLIM);

        if((s=strtok(buf," "))==NULL) return 0;
        TLCt.x=atol(s);

        if((s=strtok(NULL," "))==NULL) return 0;
        TLCt.y=atol(s);

        //text string
        fgets(buf,TLCLIM,rTLCfile);
        chop(buf,TLCLIM);
		strcpy(TLCt.text,buf);
        return 1;
    }
    return 0;
}

//write TLC header info
void writeH(void)
{
    static char buf[10];

    fprintf(wTLCfile,"=H\n");
    fprintf(wTLCfile,"%s\n",TLCh.name);
    fprintf(wTLCfile,"%s\n",TLCh.lv);
    fprintf(wTLCfile,"%s\n",TLCh.tv);
    fprintf(wTLCfile,"%g\n",TLCh.scale);
    fprintf(wTLCfile,"%s\n",TLCh.punit);
    fprintf(wTLCfile,"%s\n",_strdate(buf));
    fprintf(wTLCfile,"%s\n",_strtime(buf));
    fprintf(wTLCfile,"%d %ld %ld %ld %ld\n",TLCh.rank,TLCh.l,TLCh.b,TLCh.r,TLCh.t);
    fprintf(wTLCfile,"%d %d %ld %d\n",TLCh.bn,TLCh.pn,TLCh.vn,TLCh.cn);
}

//write a cell record
void writeC(void)
{
    TLCh.cn++;
    fprintf(wTLCfile,"=C\n");
    fprintf(wTLCfile,"%s\n",TLCc.name);
    fprintf(wTLCfile,"%d %ld %ld %d\n",TLCc.o,TLCc.x,TLCc.y,0);
}

//write a box record
void writeB(void)
{
    TLCh.bn++;
    fprintf(wTLCfile,"=B\n");
    fprintf(wTLCfile,"%d %ld %ld %ld %ld\n",TLCb.lyr,TLCb.l,TLCb.b,TLCb.r,TLCb.t);
}

//write a path record
void writeP(void)
{
	int n=0;	//coords start at 0

    TLCh.pn++;
    TLCh.vn+=TLCp.nv;
    fprintf(wTLCfile,"=P\n");
    fprintf(wTLCfile,"%d %ld %d\n",TLCp.lyr,TLCp.w,TLCp.nv);

    //write 5 vertex/line
    while(n<TLCp.nv){
        fprintf(wTLCfile,"%ld %ld ",TLCp.x[n],TLCp.y[n]);
        n++;
        if(!(n % 5)) fprintf(wTLCfile,"\n");
    }
    if((n % 5)) fprintf(wTLCfile,"\n");
}

//write a text record
void writeT(void)
{
	int n=3;

    TLCh.pn++;
    TLCh.vn+=TLCt.nv;
    fprintf(wTLCfile,"=T\n");
    fprintf(wTLCfile,"%d %ld %d %d\n",TLCt.lyr,TLCt.siz,TLCt.nv,TLCt.o);
	fprintf(wTLCfile,"%ld %ld\n",TLCt.x,TLCt.y);

	//keyword strings
	if(!strcmp("=H",TLCt.text) ||
		!strcmp("=C",TLCt.text) ||
		!strcmp("=B",TLCt.text) ||
		!strcmp("=P",TLCt.text) ||
		!strcmp("=T",TLCt.text)){
		while(n-- >0) TLCt.text[n+1]=TLCt.text[n];
		TLCt.text[0]=' ';
	}
	fprintf(wTLCfile,"%s\n",TLCt.text);
}

//****************
//misc. functions
//****************

//open any TLC for read
int ropenTLC(char *name)
{
    static char buf[13];

    sprintf(buf,"%s.TLC",name);
    if((rTLCfile=fopen(buf,"rt"))==NULL)
        return 0;
    return 1;
}

//open any TLC for write
int wopenTLC(char *name)
{
    static char buf[13];

    sprintf(buf,"%s.TLC",name);
    if((wTLCfile=fopen(buf,"wt"))==NULL)
        return 0;
    return 1;
}

//load cell ref file
//return 0 if missing, cells read if successful
int LoadREF(void)
{
    FILE *ref;
    int n=1;
    static char inbuf[80];

    if((ref=fopen("CELLS.REF","rt"))==NULL)
        return 0;

    if(atoi(fgets(inbuf,80,ref))>CELMAX)
        return 0;

    while(!feof(ref) && n<CELMAX){
        fgets(inbuf,80,ref);
        strcpy(TLCname[n],chop(inbuf,NAMLEN));
        fgets(inbuf,80,ref);
        n++;
    }
    fclose(ref);
    return --n;
}

//limit and remove NL
char *chop(char *s,int c)
{
    int l;

    l=strlen(s);
    if(s[l-1]=='\n')
        s[l-1]=0;        //remove '\n'
    s[c]=0;              //keep c chars
    return s;            //return pointer
}

//check for key abort
int abrtCk(void)
{
    int key=0;

    if(kbhit()){
        key=getch();
        if(key==0x1B) return 1;
        if(key==0 || key==0xE0) getch();
        return 0;
    }
    return 0;
}


void Help(void)
{
printf("This program moves or copies the layers of boxes, poly and text in\n");
printf("TLC files. It acts only on the TLC files in the current directory.\n");
printf("It reads the CELLS.REF file in the directory and acts on the TLC files\n");
printf("that it finds in that file.\n\n");
printf("   Command line :  mclayer m[c] oldlayer newlayer\n\n");
}
